package com.github.itzilly.sbt.features

import com.github.itzilly.sbt.renderer.RenderFuncs
import net.minecraft.client.Minecraft
import net.minecraft.init.Blocks
import net.minecraft.util.BlockPos
import net.minecraftforge.client.event.RenderWorldLastEvent
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent

enum class GemstoneColors(val color: Int) {
    RUBY(0xFF5555),     // Red
    JADE(0x00AA00),     // Lime
    AMBER(0xFFAA00),    // Orange
    SAPPHIRE(0x55FFFF), // Light Blue
    AMETHYST(0xAA00AA), // Purple
    TOPAZ(0xFFFF55),    // Yellow
    JASPER(0xFF55FF)    // Magenta
}

object GemstoneFinder {
    private var gemstoneList: ArrayList<GemstoneBlock> = ArrayList()

    fun clear() {
        gemstoneList.clear()
    }

    fun find() {
        val playerPos = Minecraft.getMinecraft().thePlayer.position

        val minBounds = BlockPos(
            playerPos.x - 32,
            maxOf(playerPos.y - 32, 31),
            playerPos.z - 32
        )

        val maxBounds = BlockPos(
            playerPos.x + 32,
            minOf(playerPos.y + 32, 188),
            playerPos.z + 32
        )
        val world = Minecraft.getMinecraft().thePlayer.worldObj

        for (chunkX in minBounds.x / 16..maxBounds.x / 16) {
            for (chunkZ in minBounds.z / 16..maxBounds.z / 16) {
                val chunk = world.getChunkFromChunkCoords(chunkX, chunkZ)

                for (x in 0..15) {
                    for (y in minBounds.y..maxBounds.y) {
                        for (z in 0..15) {
                            val worldX = chunk.xPosition * 16 + x
                            val worldZ = chunk.zPosition * 16 + z

                            if (worldX in minBounds.x..maxBounds.x && worldZ in minBounds.z..maxBounds.z) {
                                val blockPos = BlockPos(worldX, y, worldZ)
                                val blockState = chunk.getBlockState(blockPos)
                                val block = blockState.block
                                if (block == Blocks.air) continue
                                val meta = block.getMetaFromState(blockState)

                                val gemstoneColor = when {
                                    block === Blocks.stained_glass || block === Blocks.stained_glass_pane -> {
                                        when (meta) {
                                            1 -> GemstoneColors.AMBER  // Orange
                                            2 -> GemstoneColors.JASPER  // Magenta
                                            3 -> GemstoneColors.SAPPHIRE  // Light Blue
                                            4 -> GemstoneColors.TOPAZ  // Yellow
                                            5 -> GemstoneColors.JADE  // Lime
                                            14 -> GemstoneColors.RUBY  // Red
                                            10 -> GemstoneColors.AMETHYST  // Purple
                                            else -> null
                                        }
                                    }
                                    else -> null
                                }

                                if (gemstoneColor != null && !gemstonesList(blockPos.x, blockPos.y, blockPos.z)) {
                                    gemstoneList.add(GemstoneBlock(blockPos.x, blockPos.y, blockPos.z, gemstoneColor))
                                }
                            }
                        }
                    }
                }
            }
        }
        mergeAdjacentBlocks()
    }

    private fun mergeAdjacentBlocks() {
        val mergedList = ArrayList<GemstoneBlock>()

        while (gemstoneList.isNotEmpty()) {
            val firstBlock = gemstoneList.removeAt(0)
            val group = mutableListOf(firstBlock)

            val queue = ArrayDeque<GemstoneBlock>()
            queue.add(firstBlock)

            while (queue.isNotEmpty()) {
                val current = queue.removeFirst()
                val neighbors = gemstoneList.filter {
                    it.color == current.color && isAdjacent(it, current)
                }
                group.addAll(neighbors)
                gemstoneList.removeAll(neighbors)
                queue.addAll(neighbors)
            }

            mergedList.add(mergeGroup(group))
        }

        gemstoneList = mergedList
    }

    private fun isAdjacent(a: GemstoneBlock, b: GemstoneBlock): Boolean {
        return (a.x == b.x && a.y == b.y && (a.z == b.z + 1 || a.z == b.z - 1)) ||
                (a.x == b.x && (a.y == b.y + 1 || a.y == b.y - 1) && a.z == b.z) ||
                ((a.x == b.x + 1 || a.x == b.x - 1) && a.y == b.y && a.z == b.z)
    }

    private fun mergeGroup(group: List<GemstoneBlock>): GemstoneBlock {
        val minX = group.minOf { it.x }
        val minY = group.minOf { it.y }
        val minZ = group.minOf { it.z }
        val maxX = group.maxOf { it.x }
        val maxY = group.maxOf { it.y }
        val maxZ = group.maxOf { it.z }
        return GemstoneBlock(minX, minY, minZ, group.first().color, maxX, maxY, maxZ)
    }

    @SubscribeEvent
    fun renderBlockOverlay(event: RenderWorldLastEvent) {
        val player = Minecraft.getMinecraft().thePlayer

        for (gemstone: GemstoneBlock in gemstoneList) {
            val startPos = BlockPos(gemstone.x, gemstone.y, gemstone.z)
            val endPos = BlockPos(gemstone.maxX, gemstone.maxY, gemstone.maxZ)
            RenderFuncs.drawBlockOutline(player, startPos, endPos, event.partialTicks, gemstone.color.color)
        }
    }

    data class GemstoneBlock(
        val x: Int,
        val y: Int,
        val z: Int,
        val color: GemstoneColors,
        val maxX: Int = x,
        val maxY: Int = y,
        val maxZ: Int = z
    )

    fun gemstonesList(x: Int, y: Int, z: Int): Boolean {
        return gemstoneList.any { it.x == x && it.y == y && it.z == z }
    }
}